"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const v4_1 = tslib_1.__importDefault(require("uuid/v4"));
const Logger_1 = tslib_1.__importDefault(require("../../common/Logger"));
const constants_1 = require("../../common/utils/constants");
const BlobStorageContext_1 = tslib_1.__importDefault(require("../context/BlobStorageContext"));
const StorageErrorFactory_1 = tslib_1.__importDefault(require("../errors/StorageErrorFactory"));
const constants_2 = require("../utils/constants");
const utils_1 = require("../utils/utils");
/**
 * A middleware extract related blob service context.
 *
 * @export
 * @param {Request} req An express compatible Request object
 * @param {Response} res An express compatible Response object
 * @param {NextFunction} next An express middleware next callback
 */
function blobStorageContextMiddleware(req, res, next) {
    // Set server header in every Azurite response
    res.setHeader(constants_2.HeaderConstants.SERVER, `Azurite-Blob/${constants_2.VERSION}`);
    const requestID = v4_1.default();
    const apiVersion = req.header(constants_2.HeaderConstants.X_MS_VERSION);
    if (apiVersion !== undefined) {
        utils_1.checkApiVersion(apiVersion, constants_2.ValidAPIVersions, requestID);
    }
    const blobContext = new BlobStorageContext_1.default(res.locals, constants_2.DEFAULT_CONTEXT_PATH);
    blobContext.startTime = new Date();
    blobContext.xMsRequestID = requestID;
    Logger_1.default.info(`BlobStorageContextMiddleware: RequestMethod=${req.method} RequestURL=${req.protocol}://${req.hostname}${req.url} RequestHeaders:${JSON.stringify(req.headers)} ClientIP=${req.ip} Protocol=${req.protocol} HTTPVersion=${req.httpVersion}`, requestID);
    const [account, container, blob, isSecondary] = extractStoragePartsFromPath(req.hostname, req.path);
    blobContext.account = account;
    blobContext.container = container;
    blobContext.blob = blob;
    blobContext.isSecondary = isSecondary;
    // Emulator's URL pattern is like http://hostname[:port]/account/container
    // (or, alternatively, http[s]://account.localhost[:port]/container)
    // Create a router to exclude account name from req.path, as url path in swagger doesn't include account
    // Exclude account name from req.path for dispatchMiddleware
    blobContext.dispatchPattern = container
        ? blob
            ? `/container/blob`
            : `/container`
        : "/";
    blobContext.authenticationPath = req.path;
    if (isSecondary) {
        const pos = blobContext.authenticationPath.search(constants_2.SECONDARY_SUFFIX);
        blobContext.authenticationPath =
            blobContext.authenticationPath.substr(0, pos) +
                blobContext.authenticationPath.substr(pos + constants_2.SECONDARY_SUFFIX.length);
    }
    if (!account) {
        const handlerError = StorageErrorFactory_1.default.getInvalidQueryParameterValue(requestID);
        Logger_1.default.error(`BlobStorageContextMiddleware: BlobStorageContextMiddleware: ${handlerError.message}`, requestID);
        return next(handlerError);
    }
    Logger_1.default.info(`BlobStorageContextMiddleware: Account=${account} Container=${container} Blob=${blob}`, requestID);
    next();
}
exports.default = blobStorageContextMiddleware;
/**
 * Extract storage account, container, and blob from URL path.
 *
 * @param {string} path
 * @returns {([string | undefined, string | undefined, string | undefined, boolean | undefined])}
 */
function extractStoragePartsFromPath(hostname, path) {
    let account;
    let container;
    let blob;
    let isSecondary = false;
    const decodedPath = decodeURIComponent(path);
    const normalizedPath = decodedPath.startsWith("/")
        ? decodedPath.substr(1)
        : decodedPath; // Remove starting "/"
    const parts = normalizedPath.split("/");
    let urlPartIndex = 0;
    if (hostname.endsWith(constants_1.PRODUCTION_STYLE_URL_HOSTNAME)) {
        account = hostname.substring(0, hostname.length - constants_1.PRODUCTION_STYLE_URL_HOSTNAME.length);
    }
    else {
        account = parts[urlPartIndex++];
    }
    container = parts[urlPartIndex++];
    blob = parts
        .slice(urlPartIndex++)
        .join("/")
        .replace(/\\/g, "/"); // Azure Storage Server will replace "\" with "/" in the blob names
    if (account.endsWith(constants_2.SECONDARY_SUFFIX)) {
        account = account.substr(0, account.length - constants_2.SECONDARY_SUFFIX.length);
        isSecondary = true;
    }
    return [account, container, blob, isSecondary];
}
exports.extractStoragePartsFromPath = extractStoragePartsFromPath;
//# sourceMappingURL=blobStorageContext.middleware.js.map