"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const crypto_1 = require("crypto");
const etag_1 = tslib_1.__importDefault(require("etag"));
const fs_1 = require("fs");
const url_1 = require("url");
const xml2js = tslib_1.__importStar(require("xml2js"));
const StorageErrorFactory_1 = tslib_1.__importDefault(require("../errors/StorageErrorFactory"));
/**
 * Generates a hash signature for an HTTP request or for a SAS.
 *
 * @param {string} stringToSign
 * @param {key} key
 * @returns {string}
 */
function computeHMACSHA256(stringToSign, key) {
    return crypto_1.createHmac("sha256", key)
        .update(stringToSign, "utf8")
        .digest("base64");
}
exports.computeHMACSHA256 = computeHMACSHA256;
function checkApiVersion(inputApiVersion, validApiVersions, requestId) {
    if (!validApiVersions.includes(inputApiVersion)) {
        throw StorageErrorFactory_1.default.getInvalidHeaderValue(requestId, {
            HeaderName: "x-ms-version",
            HeaderValue: inputApiVersion
        });
    }
}
exports.checkApiVersion = checkApiVersion;
/**
 * Rounds a date off to seconds.
 *
 * @export
 * @param {Date} date
 * @param {boolean} [withMilliseconds=true] If true, YYYY-MM-DDThh:mm:ss.fffffffZ will be returned;
 *                                          If false, YYYY-MM-DDThh:mm:ssZ will be returned.
 * @returns {string} Date string in ISO8061 format, with or without 7 milliseconds component
 */
function truncatedISO8061Date(date, withMilliseconds = true) {
    // Date.toISOString() will return like "2018-10-29T06:34:36.139Z"
    const dateString = date.toISOString();
    return withMilliseconds
        ? dateString.substring(0, dateString.length - 1) + "0000" + "Z"
        : dateString.substring(0, dateString.length - 5) + "Z";
}
exports.truncatedISO8061Date = truncatedISO8061Date;
// TODO: Align eTag with Azure Storage Service
function newEtag() {
    return etag_1.default(`${new Date().getTime()}`);
}
exports.newEtag = newEtag;
async function streamToLocalFile(stream, path) {
    return new Promise((resolve, reject) => {
        const writeStream = fs_1.createWriteStream(path);
        stream
            .on("error", reject)
            // .on("end", resolve)
            .pipe(writeStream)
            .on("close", resolve)
            .on("error", reject);
    });
}
exports.streamToLocalFile = streamToLocalFile;
async function getMD5FromStream(stream) {
    const hash = crypto_1.createHash("md5");
    return new Promise((resolve, reject) => {
        stream
            .on("data", data => {
            hash.update(data);
        })
            .on("end", () => {
            resolve(hash.digest());
        })
            .on("error", reject);
    });
}
exports.getMD5FromStream = getMD5FromStream;
async function getMD5FromString(text) {
    return crypto_1.createHash("md5")
        .update(text)
        .digest();
}
exports.getMD5FromString = getMD5FromString;
/**
 * Get URL query key value pairs from an URL string.
 *
 * @export
 * @param {string} url
 * @returns {{[key: string]: string}}
 */
function getURLQueries(url) {
    let queryString = url_1.parse(url).query;
    if (!queryString) {
        return {};
    }
    queryString = queryString.trim();
    queryString = queryString.startsWith("?")
        ? queryString.substr(1)
        : queryString;
    let querySubStrings = queryString.split("&");
    querySubStrings = querySubStrings.filter((value) => {
        const indexOfEqual = value.indexOf("=");
        const lastIndexOfEqual = value.lastIndexOf("=");
        return indexOfEqual > 0 && indexOfEqual === lastIndexOfEqual;
    });
    const queries = {};
    for (const querySubString of querySubStrings) {
        const splitResults = querySubString.split("=");
        const key = splitResults[0];
        const value = splitResults[1];
        queries[key] = value;
    }
    return queries;
}
exports.getURLQueries = getURLQueries;
/**
 * Default range value [0, Infinite] will be returned if all parameters not provided.
 *
 * @export
 * @param {string} [rangeHeaderValue]
 * @param {string} [xMsRangeHeaderValue]
 * @returns {[number, number]}
 */
function deserializeRangeHeader(rangeHeaderValue, xMsRangeHeaderValue) {
    const range = xMsRangeHeaderValue || rangeHeaderValue;
    if (!range) {
        return [0, Infinity];
    }
    let parts = range.split("=");
    if (parts === undefined || parts.length !== 2) {
        throw new RangeError(`deserializeRangeHeader: raw range value ${range} is wrong.`);
    }
    parts = parts[1].split("-");
    if (parts === undefined || parts.length < 1 || parts.length > 2) {
        throw new RangeError(`deserializeRangeHeader: raw range value ${range} is wrong.`);
    }
    const startInclusive = parseInt(parts[0], 10);
    let endInclusive = Infinity;
    if (parts.length > 1 && parts[1] !== "") {
        endInclusive = parseInt(parts[1], 10);
    }
    if (startInclusive > endInclusive) {
        throw new RangeError(`deserializeRangeHeader: raw range value ${range} is wrong.`);
    }
    return [startInclusive, endInclusive];
}
exports.deserializeRangeHeader = deserializeRangeHeader;
/**
 * Deserialize range header into valid page ranges.
 * For example, "bytes=0-1023" will return [0, 1023].
 *
 * Default range value [0, Infinite] will be returned if all parameters not provided.
 *
 * @private
 * @param {string} [rangeHeaderValue]
 * @param {string} [xMsRangeHeaderValue]
 * @returns {([number, number] | undefined)}
 */
/**
 * validate the input name for queue or container.
 * @see https://docs.microsoft.com/en-us/rest/api/storageservices/naming-queues-and-metadata
 *
 * @export
 * @param {string} name
 * @returns {nameValidateCode} //0 for valid, 1 for outOfRange, 2 for invalid.
 */
function isValidName(name) {
    if (name === "") {
        return nameValidateCode.invalidName;
    }
    if (name.length < 3 || name.length > 63) {
        return nameValidateCode.outOfRange;
    }
    if (name.split("-").indexOf("") !== -1) {
        return nameValidateCode.invalidName;
    }
    const reg = new RegExp("^[0-9|a-z|-]*$");
    if (reg.test(name)) {
        return nameValidateCode.valid;
    }
    return nameValidateCode.invalidName;
}
exports.isValidName = isValidName;
// The code to indicate the validation result.
var nameValidateCode;
(function (nameValidateCode) {
    nameValidateCode[nameValidateCode["valid"] = 0] = "valid";
    nameValidateCode[nameValidateCode["outOfRange"] = 1] = "outOfRange";
    nameValidateCode[nameValidateCode["invalidUri"] = 2] = "invalidUri";
    nameValidateCode[nameValidateCode["invalidName"] = 3] = "invalidName";
})(nameValidateCode = exports.nameValidateCode || (exports.nameValidateCode = {}));
/**
 * Generate a random code with given length
 *
 * @public
 * @param {number} len
 * @returns {string}
 * @memberof LokiQueueDataStore
 */
function randomValueHex(len) {
    return crypto_1.randomBytes(Math.ceil(len / 2))
        .toString("hex") // convert to hexadecimal format
        .slice(0, len); // return required number of characters
}
exports.randomValueHex = randomValueHex;
/**
 * Generate the popreceipt for a get messages request.
 *
 * @public
 * @param {Date} requestDate
 * @returns {string}
 * @memberof LokiQueueDataStore
 */
function getPopReceipt(requestDate) {
    const encodedStr = requestDate
        .toUTCString()
        .split(" ")
        .slice(1, 5)
        .join("") + randomValueHex(4);
    return Buffer.from(encodedStr).toString("base64");
}
exports.getPopReceipt = getPopReceipt;
/**
 * Read the text from a readStream to a string.
 *
 * @export
 * @param {NodeJS.ReadableStream} data
 * @returns {Promise<string>}
 */
async function readStreamToString(data) {
    return new Promise((resolve, reject) => {
        let res = "";
        data
            .on("readable", () => {
            let chunk;
            chunk = data.read();
            if (chunk) {
                res += chunk.toString();
            }
        })
            .on("end", () => {
            resolve(res);
        })
            .on("error", reject);
    });
}
exports.readStreamToString = readStreamToString;
/**
 * Get the byte size of a string in UTF8.
 *
 * @public
 * @param {Date} requestDate
 * @returns {string}
 * @memberof LokiQueueDataStore
 */
function getUTF8ByteSize(text) {
    return Buffer.from(text, "utf8").length;
}
exports.getUTF8ByteSize = getUTF8ByteSize;
/**
 * Retrive the value from XML body without ignoring the empty characters.
 *
 * @export
 * @param {string} param
 * @param {boolean} [explicitChildrenWithOrder=false]
 * @returns {Promise<any>}
 */
function parseXMLwithEmpty(param, explicitChildrenWithOrder = false) {
    const xmlParser = new xml2js.Parser({
        explicitArray: false,
        explicitCharkey: false,
        explicitRoot: false,
        preserveChildrenOrder: explicitChildrenWithOrder,
        explicitChildren: explicitChildrenWithOrder
    });
    return new Promise((resolve, reject) => {
        xmlParser.parseString(param, (err, res) => {
            if (err) {
                reject(err);
            }
            else {
                resolve(res);
            }
        });
    });
}
exports.parseXMLwithEmpty = parseXMLwithEmpty;
//# sourceMappingURL=utils.js.map