import IGCExtentProvider from "../../common/IGCExtentProvider";
import * as Models from "../generated/artifacts/models";
import Context from "../generated/Context";
import IBlobMetadataStore, { AcquireBlobLeaseResponse, AcquireContainerLeaseResponse, BlobId, BlobModel, BlockModel, BreakBlobLeaseResponse, BreakContainerLeaseResponse, ChangeBlobLeaseResponse, ChangeContainerLeaseResponse, ContainerModel, CreateSnapshotResponse, GetBlobPropertiesRes, GetContainerAccessPolicyResponse, GetContainerPropertiesResponse, GetPageRangeResponse, IContainerMetadata, IExtentChunk, ReleaseBlobLeaseResponse, ReleaseContainerLeaseResponse, RenewBlobLeaseResponse, RenewContainerLeaseResponse, ServicePropertiesModel, SetContainerAccessPolicyOptions } from "./IBlobMetadataStore";
/**
 * This is a metadata source implementation for blob based on loki DB.
 *
 * Notice that, following design is for emulator purpose only, and doesn't design for best performance.
 * We may want to optimize the persistency layer performance in the future. Such as by distributing metadata
 * into different collections, or make binary payload write as an append-only pattern.
 *
 * Loki DB includes following collections and documents:
 *
 * -- SERVICE_PROPERTIES_COLLECTION // Collection contains service properties
 *                                  // Default collection name is $SERVICES_COLLECTION$
 *                                  // Each document maps to 1 account blob service
 *                                  // Unique document properties: accountName
 * -- CONTAINERS_COLLECTION  // Collection contains all containers
 *                           // Default collection name is $CONTAINERS_COLLECTION$
 *                           // Each document maps to 1 container
 *                           // Unique document properties: accountName, (container)name
 * -- BLOBS_COLLECTION       // Collection contains all blobs
 *                           // Default collection name is $BLOBS_COLLECTION$
 *                           // Each document maps to a blob
 *                           // Unique document properties: accountName, containerName, (blob)name, snapshot
 * -- BLOCKS_COLLECTION      // Block blob blocks collection includes all UNCOMMITTED blocks
 *                           // Unique document properties: accountName, containerName, blobName, name, isCommitted
 *
 * @export
 * @class LokiBlobMetadataStore
 */
export default class LokiBlobMetadataStore implements IBlobMetadataStore, IGCExtentProvider {
    readonly lokiDBPath: string;
    private readonly db;
    private initialized;
    private closed;
    private readonly SERVICES_COLLECTION;
    private readonly CONTAINERS_COLLECTION;
    private readonly BLOBS_COLLECTION;
    private readonly BLOCKS_COLLECTION;
    private readonly pageBlobRangesManager;
    constructor(lokiDBPath: string);
    isInitialized(): boolean;
    isClosed(): boolean;
    init(): Promise<void>;
    /**
     * Close loki DB.
     *
     * @returns {Promise<void>}
     * @memberof LokiBlobMetadataStore
     */
    close(): Promise<void>;
    /**
     * Clean LokiBlobMetadataStore.
     *
     * @returns {Promise<void>}
     * @memberof LokiBlobMetadataStore
     */
    clean(): Promise<void>;
    iteratorExtents(): AsyncIterator<string[]>;
    /**
     * Update blob service properties. Create service properties if not exists in persistency layer.
     *
     * TODO: Account's service property should be created when storage account is created or metadata
     * storage initialization. This method should only be responsible for updating existing record.
     * In this way, we can reduce one I/O call to get account properties.
     *
     * @param {ServicePropertiesModel} serviceProperties
     * @returns {Promise<ServicePropertiesModel>} undefined properties will be ignored during properties setup
     * @memberof LokiBlobMetadataStore
     */
    setServiceProperties(context: Context, serviceProperties: ServicePropertiesModel): Promise<ServicePropertiesModel>;
    /**
     * Get service properties for specific storage account.
     *
     * @param {string} account
     * @returns {Promise<ServicePropertiesModel | undefined>}
     * @memberof LokiBlobMetadataStore
     */
    getServiceProperties(context: Context, account: string): Promise<ServicePropertiesModel | undefined>;
    /**
     * List containers with query conditions specified.
     *
     * @param {string} account
     * @param {string} [prefix=""]
     * @param {number} [maxResults=5000]
     * @param {string} [marker=""]
     * @returns {(Promise<[ContainerModel[], string | undefined]>)}
     * @memberof LokiBlobMetadataStore
     */
    listContainers(context: Context, account: string, prefix?: string, maxResults?: number, marker?: string): Promise<[ContainerModel[], string | undefined]>;
    /**
     * Create a container.
     *
     * @param {ContainerModel} container
     * @returns {Promise<ContainerModel>}
     * @memberof LokiBlobMetadataStore
     */
    createContainer(context: Context, container: ContainerModel): Promise<ContainerModel>;
    /**
     * Get container properties.
     *
     * @param {string} account
     * @param {string} container
     * @param {Context} context
     * @param {Models.LeaseAccessConditions} [leaseAccessConditions]
     * @returns {Promise<GetContainerPropertiesResponse>}
     * @memberof LokiBlobMetadataStore
     */
    getContainerProperties(context: Context, account: string, container: string, leaseAccessConditions?: Models.LeaseAccessConditions): Promise<GetContainerPropertiesResponse>;
    /**
     * Delete container item if exists from persistency layer.
     *
     * Loki based implementation will delete container documents from Containers collection,
     * blob documents from Blobs collection, and blocks documents from Blocks collection immediately.
     *
     * Persisted extents data will be deleted by GC.
     *
     * @param {string} account
     * @param {string} container
     * @param {Context} context
     * @param {Models.ContainerDeleteMethodOptionalParams} [options]
     * @returns {Promise<void>}
     * @memberof LokiBlobMetadataStore
     */
    deleteContainer(context: Context, account: string, container: string, options?: Models.ContainerDeleteMethodOptionalParams): Promise<void>;
    /**
     * Set container metadata.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {Date} lastModified
     * @param {string} etag
     * @param {IContainerMetadata} [metadata]
     * @param {Models.LeaseAccessConditions} [leaseAccessConditions]
     * @param {Models.ModifiedAccessConditions} [modifiedAccessConditions]
     * @returns {Promise<void>}
     * @memberof LokiBlobMetadataStore
     */
    setContainerMetadata(context: Context, account: string, container: string, lastModified: Date, etag: string, metadata?: IContainerMetadata, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<void>;
    /**
     * Get container access policy.
     *
     * @param {string} account
     * @param {string} container
     * @param {Context} context
     * @param {Models.LeaseAccessConditions} [leaseAccessConditions]
     * @returns {Promise<GetContainerAccessPolicyResponse>}
     * @memberof LokiBlobMetadataStore
     */
    getContainerACL(context: Context, account: string, container: string, leaseAccessConditions?: Models.LeaseAccessConditions): Promise<GetContainerAccessPolicyResponse>;
    /**
     * Set container access policy.
     *
     * @param {string} account
     * @param {string} container
     * @param {SetContainerAccessPolicyOptions} setAclModel
     * @param {Context} context
     * @returns {Promise<void>}
     * @memberof LokiBlobMetadataStore
     */
    setContainerACL(context: Context, account: string, container: string, setAclModel: SetContainerAccessPolicyOptions): Promise<void>;
    /**
     * Acquire container lease.
     *
     * @param {string} account
     * @param {string} container
     * @param {Models.ContainerAcquireLeaseOptionalParams} options
     * @param {Context} context
     * @returns {Promise<AcquireContainerLeaseResponse>}
     * @memberof LokiBlobMetadataStore
     */
    acquireContainerLease(context: Context, account: string, container: string, options: Models.ContainerAcquireLeaseOptionalParams): Promise<AcquireContainerLeaseResponse>;
    /**
     * Release container lease.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} leaseId
     * @param {Models.ContainerReleaseLeaseOptionalParams} [options={}]
     * @returns {Promise<ReleaseContainerLeaseResponse>}
     * @memberof LokiBlobMetadataStore
     */
    releaseContainerLease(context: Context, account: string, container: string, leaseId: string, options?: Models.ContainerReleaseLeaseOptionalParams): Promise<ReleaseContainerLeaseResponse>;
    /**
     * Renew container lease.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} leaseId
     * @param {Models.ContainerRenewLeaseOptionalParams} [options={}]
     * @returns {Promise<RenewContainerLeaseResponse>}
     * @memberof LokiBlobMetadataStore
     */
    renewContainerLease(context: Context, account: string, container: string, leaseId: string, options?: Models.ContainerRenewLeaseOptionalParams): Promise<RenewContainerLeaseResponse>;
    /**
     * Break container lease.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {(number | undefined)} breakPeriod
     * @param {Models.ContainerBreakLeaseOptionalParams} [options={}]
     * @returns {Promise<BreakContainerLeaseResponse>}
     * @memberof LokiBlobMetadataStore
     */
    breakContainerLease(context: Context, account: string, container: string, breakPeriod: number | undefined, options?: Models.ContainerBreakLeaseOptionalParams): Promise<BreakContainerLeaseResponse>;
    /**
     * Change container lease.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} leaseId
     * @param {string} proposedLeaseId
     * @param {Models.ContainerChangeLeaseOptionalParams} [options={}]
     * @returns {Promise<ChangeContainerLeaseResponse>}
     * @memberof LokiBlobMetadataStore
     */
    changeContainerLease(context: Context, account: string, container: string, leaseId: string, proposedLeaseId: string, options?: Models.ContainerChangeLeaseOptionalParams): Promise<ChangeContainerLeaseResponse>;
    /**
     * Check the existence of a container.
     *
     * @param {string} account
     * @param {string} container
     * @param {Context} [context]
     * @returns {Promise<void>}
     * @memberof LokiBlobMetadataStore
     */
    checkContainerExist(context: Context, account: string, container: string): Promise<void>;
    listBlobs(context: Context, account: string, container: string, blob?: string, prefix?: string, maxResults?: number, marker?: string, includeSnapshots?: boolean): Promise<[BlobModel[], string | undefined]>;
    listAllBlobs(maxResults?: number, marker?: string, includeSnapshots?: boolean): Promise<[BlobModel[], string | undefined]>;
    /**
     * Create blob item in persistency layer. Will replace if blob exists.
     *
     * @param {Context} context
     * @param {BlobModel} blob
     * @param {Models.LeaseAccessConditions} [leaseAccessConditions]
     * @param {Models.ModifiedAccessConditions} [modifiedAccessConditions]
     * @returns {Promise<void>}
     * @memberof LokiBlobMetadataStore
     */
    createBlob(context: Context, blob: BlobModel, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<void>;
    /**
     * Create snapshot.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {Models.LeaseAccessConditions} [leaseAccessConditions] Optional. Will validate lease if provided
     * @returns {Promise<CreateSnapshotResponse>}
     * @memberof LokiBlobMetadataStore
     */
    createSnapshot(context: Context, account: string, container: string, blob: string, leaseAccessConditions?: Models.LeaseAccessConditions, metadata?: Models.BlobMetadata, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<CreateSnapshotResponse>;
    /**
     * Gets a blob item from persistency layer by container name and blob name.
     * Will return block list or page list as well for downloading.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {string} [snapshot=""]
     * @param {Models.LeaseAccessConditions} [leaseAccessConditions]
     * @param {Models.ModifiedAccessConditions} [modifiedAccessConditions]
     * @returns {Promise<BlobModel>}
     * @memberof LokiBlobMetadataStore
     */
    downloadBlob(context: Context, account: string, container: string, blob: string, snapshot?: string, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<BlobModel>;
    /**
     * Gets a blob item from persistency layer by container name and blob name.
     * Will return block list or page list as well for downloading.
     *
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {string} [snapshot]
     * @returns {(Promise<BlobModel | undefined>)}
     * @memberof LokiBlobMetadataStore
     */
    getBlob(context: Context, account: string, container: string, blob: string, snapshot?: string): Promise<BlobModel | undefined>;
    /**
     * Get blob properties.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {string} [snapshot=""]
     * @param {(Models.LeaseAccessConditions | undefined)} leaseAccessConditions
     * @param {Models.ModifiedAccessConditions} [modifiedAccessConditions]
     * @returns {Promise<GetBlobPropertiesRes>}
     * @memberof LokiBlobMetadataStore
     */
    getBlobProperties(context: Context, account: string, container: string, blob: string, snapshot: string | undefined, leaseAccessConditions: Models.LeaseAccessConditions | undefined, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<GetBlobPropertiesRes>;
    /**
     * Delete blob or its snapshots.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {Models.BlobDeleteMethodOptionalParams} options
     * @returns {Promise<void>}
     * @memberof LokiBlobMetadataStore
     */
    deleteBlob(context: Context, account: string, container: string, blob: string, options: Models.BlobDeleteMethodOptionalParams): Promise<void>;
    /**
     * Set blob HTTP headers.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {(Models.LeaseAccessConditions | undefined)} leaseAccessConditions
     * @param {(Models.BlobHTTPHeaders | undefined)} blobHTTPHeaders
     * @param {Models.ModifiedAccessConditions} [modifiedAccessConditions]
     * @returns {Promise<Models.BlobProperties>}
     * @memberof LokiBlobMetadataStore
     */
    setBlobHTTPHeaders(context: Context, account: string, container: string, blob: string, leaseAccessConditions: Models.LeaseAccessConditions | undefined, blobHTTPHeaders: Models.BlobHTTPHeaders | undefined, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<Models.BlobProperties>;
    /**
     * Set blob metadata.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {(Models.LeaseAccessConditions | undefined)} leaseAccessConditions
     * @param {(Models.BlobMetadata | undefined)} metadata
     * @param {Models.ModifiedAccessConditions} [modifiedAccessConditions]
     * @returns {Promise<Models.BlobProperties>}
     * @memberof LokiBlobMetadataStore
     */
    setBlobMetadata(context: Context, account: string, container: string, blob: string, leaseAccessConditions: Models.LeaseAccessConditions | undefined, metadata: Models.BlobMetadata | undefined, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<Models.BlobProperties>;
    /**
     * Acquire blob lease.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {number} duration
     * @param {string} [proposedLeaseId]
     * @param {Models.BlobAcquireLeaseOptionalParams} [options={}]
     * @returns {Promise<AcquireBlobLeaseResponse>}
     * @memberof LokiBlobMetadataStore
     */
    acquireBlobLease(context: Context, account: string, container: string, blob: string, duration: number, proposedLeaseId?: string, options?: Models.BlobAcquireLeaseOptionalParams): Promise<AcquireBlobLeaseResponse>;
    /**
     * Release blob.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {string} leaseId
     * @param {Models.BlobReleaseLeaseOptionalParams} [options={}]
     * @returns {Promise<ReleaseBlobLeaseResponse>}
     * @memberof LokiBlobMetadataStore
     */
    releaseBlobLease(context: Context, account: string, container: string, blob: string, leaseId: string, options?: Models.BlobReleaseLeaseOptionalParams): Promise<ReleaseBlobLeaseResponse>;
    /**
     * Renew blob lease.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {string} leaseId
     * @param {Models.BlobRenewLeaseOptionalParams} [options={}]
     * @returns {Promise<RenewBlobLeaseResponse>}
     * @memberof LokiBlobMetadataStore
     */
    renewBlobLease(context: Context, account: string, container: string, blob: string, leaseId: string, options?: Models.BlobRenewLeaseOptionalParams): Promise<RenewBlobLeaseResponse>;
    /**
     * Change blob lease.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {string} leaseId
     * @param {string} proposedLeaseId
     * @param {Models.BlobChangeLeaseOptionalParams} [option={}]
     * @returns {Promise<ChangeBlobLeaseResponse>}
     * @memberof LokiBlobMetadataStore
     */
    changeBlobLease(context: Context, account: string, container: string, blob: string, leaseId: string, proposedLeaseId: string, options?: Models.BlobChangeLeaseOptionalParams): Promise<ChangeBlobLeaseResponse>;
    /**
     * Break blob lease.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {(number | undefined)} breakPeriod
     * @param {Models.BlobBreakLeaseOptionalParams} [options={}]
     * @returns {Promise<BreakBlobLeaseResponse>}
     * @memberof LokiBlobMetadataStore
     */
    breakBlobLease(context: Context, account: string, container: string, blob: string, breakPeriod: number | undefined, options?: Models.BlobBreakLeaseOptionalParams): Promise<BreakBlobLeaseResponse>;
    /**
     * Check the existence of a blob.
     *
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {string} [snapshot=""]
     * @param {Context} [context]
     * @returns {Promise<void>}
     * @memberof LokiBlobMetadataStore
     */
    checkBlobExist(context: Context, account: string, container: string, blob: string, snapshot?: string): Promise<void>;
    /**
     * Get blobType and committed status for SAS authentication.
     *
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {string} [snapshot=""]
     * @returns {(Promise<
     *     { blobType: Models.BlobType | undefined; isCommitted: boolean } | undefined
     *   >)}
     * @memberof LokiBlobMetadataStore
     */
    getBlobType(account: string, container: string, blob: string, snapshot?: string): Promise<{
        blobType: Models.BlobType | undefined;
        isCommitted: boolean;
    } | undefined>;
    /**
     * Start copy from Url.
     *
     * @param {Context} context
     * @param {BlobId} source
     * @param {BlobId} destination
     * @param {string} copySource
     * @param {(Models.BlobMetadata | undefined)} metadata
     * @param {(Models.AccessTier | undefined)} tier
     * @param {Models.BlobStartCopyFromURLOptionalParams} [leaseAccessConditions]
     * @returns {Promise<Models.BlobProperties>}
     * @memberof LokiBlobMetadataStore
     */
    startCopyFromURL(context: Context, source: BlobId, destination: BlobId, copySource: string, metadata: Models.BlobMetadata | undefined, tier: Models.AccessTier | undefined, options?: Models.BlobStartCopyFromURLOptionalParams): Promise<Models.BlobProperties>;
    /**
     * Update Tier for a blob.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {Models.AccessTier} tier
     * @param {(Models.LeaseAccessConditions | undefined)} leaseAccessConditions
     * @returns {(Promise<200 | 202>)}
     * @memberof LokiBlobMetadataStore
     */
    setTier(context: Context, account: string, container: string, blob: string, tier: Models.AccessTier, leaseAccessConditions: Models.LeaseAccessConditions | undefined): Promise<200 | 202>;
    /**
     * Update blob block item in persistency layer. Will create if block doesn't exist.
     * Will also create a uncommitted block blob.
     *
     * @param {BlockModel} block
     * @param {Context} context
     * @returns {Promise<void>}
     * @memberof LokiBlobMetadataStore
     */
    stageBlock(context: Context, block: BlockModel, leaseAccessConditions?: Models.LeaseAccessConditions): Promise<void>;
    appendBlock(context: Context, block: BlockModel, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions, appendPositionAccessConditions?: Models.AppendPositionAccessConditions): Promise<Models.BlobProperties>;
    /**
     * Commit block list for a blob.
     *
     * @param {Context} context
     * @param {BlobModel} blob
     * @param {{ blockName: string; blockCommitType: string }[]} blockList
     * @param {Models.LeaseAccessConditions} [leaseAccessConditions]
     * @param {Models.ModifiedAccessConditions} [modifiedAccessConditions]
     * @returns {Promise<void>}
     * @memberof LokiBlobMetadataStore
     */
    commitBlockList(context: Context, blob: BlobModel, blockList: {
        blockName: string;
        blockCommitType: string;
    }[], leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<void>;
    /**
     * Gets blocks list for a blob from persistency layer by account, container and blob names.
     *
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {(boolean | undefined)} isCommitted
     * @param {Context} context
     * @returns {Promise<{
     *     properties: Models.BlobProperties;
     *     uncommittedBlocks: Models.Block[];
     *     committedBlocks: Models.Block[];
     *   }>}
     * @memberof LokiBlobMetadataStore
     */
    getBlockList(context: Context, account: string, container: string, blob: string, isCommitted: boolean | undefined, leaseAccessConditions: Models.LeaseAccessConditions | undefined): Promise<{
        properties: Models.BlobProperties;
        uncommittedBlocks: Models.Block[];
        committedBlocks: Models.Block[];
    }>;
    /**
     * Upload new pages for page blob.
     *
     * @param {Context} context
     * @param {BlobModel} blob
     * @param {number} start
     * @param {number} end
     * @param {IExtentChunk} persistency
     * @param {Models.LeaseAccessConditions} [leaseAccessConditions]
     * @param {Models.ModifiedAccessConditions} [modifiedAccessConditions]
     * @param {Models.SequenceNumberAccessConditions} [sequenceNumberAccessConditions]
     * @returns {Promise<Models.BlobProperties>}
     * @memberof LokiBlobMetadataStore
     */
    uploadPages(context: Context, blob: BlobModel, start: number, end: number, persistency: IExtentChunk, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions, sequenceNumberAccessConditions?: Models.SequenceNumberAccessConditions): Promise<Models.BlobProperties>;
    /**
     * Clear range for a page blob.
     *
     * @param {Context} context
     * @param {BlobModel} blob
     * @param {number} start
     * @param {number} end
     * @param {Models.LeaseAccessConditions} [leaseAccessConditions]
     * @param {Models.ModifiedAccessConditions} [modifiedAccessConditions]
     * @param {Models.SequenceNumberAccessConditions} [sequenceNumberAccessConditions]
     * @returns {Promise<Models.BlobProperties>}
     * @memberof LokiBlobMetadataStore
     */
    clearRange(context: Context, blob: BlobModel, start: number, end: number, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions, sequenceNumberAccessConditions?: Models.SequenceNumberAccessConditions): Promise<Models.BlobProperties>;
    /**
     * Returns the list of valid page ranges for a page blob or snapshot of a page blob.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {string} [snapshot]
     * @param {Models.LeaseAccessConditions} [leaseAccessConditions]
     * @param {Models.ModifiedAccessConditions} [modifiedAccessConditions]
     * @returns {Promise<GetPageRangeResponse>}
     * @memberof LokiBlobMetadataStore
     */
    getPageRanges(context: Context, account: string, container: string, blob: string, snapshot?: string, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<GetPageRangeResponse>;
    /**
     * Resize a page blob.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {number} blobContentLength
     * @param {Models.LeaseAccessConditions} [leaseAccessConditions]
     * @param {Models.ModifiedAccessConditions} [modifiedAccessConditions]
     * @returns {Promise<Models.BlobProperties>}
     * @memberof LokiBlobMetadataStore
     */
    resizePageBlob(context: Context, account: string, container: string, blob: string, blobContentLength: number, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<Models.BlobProperties>;
    /**
     * Update the sequence number of a page blob.
     *
     * @param {Context} context
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {Models.SequenceNumberActionType} sequenceNumberAction
     * @param {(number | undefined)} blobSequenceNumber
     * @param {Models.LeaseAccessConditions} [leaseAccessConditions]
     * @param {Models.ModifiedAccessConditions} [modifiedAccessConditions]
     * @returns {Promise<Models.BlobProperties>}
     * @memberof LokiBlobMetadataStore
     */
    updateSequenceNumber(context: Context, account: string, container: string, blob: string, sequenceNumberAction: Models.SequenceNumberActionType, blobSequenceNumber: number | undefined, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<Models.BlobProperties>;
    listUncommittedBlockPersistencyChunks(marker?: string, maxResults?: number): Promise<[IExtentChunk[], string | undefined]>;
    /**
     * LokiJS will persist Uint8Array into Object.
     * This method will restore object to Uint8Array.
     *
     * @private
     * @param {*} obj
     * @returns {(Uint8Array | undefined)}
     * @memberof LokiBlobMetadataStore
     */
    private restoreUint8Array;
    /**
     * Escape a string to be used as a regex.
     *
     * @private
     * @param {string} regex
     * @returns {string}
     * @memberof LokiBlobMetadataStore
     */
    private escapeRegex;
    /**
     * Get a container document from container collection.
     * Updated lease related properties according to current time.
     *
     * @private
     * @param {string} account
     * @param {string} container
     * @param {Context} context
     * @returns {Promise<ContainerModel>}
     * @memberof LokiBlobMetadataStore
     */
    /**
     * Get a container document from container collection.
     * Updated lease related properties according to current time.
     * Will throw ContainerNotFound storage error if container doesn't exist.
     *
     * @private
     * @param {string} account
     * @param {string} container
     * @param {Context} context
     * @returns {Promise<ContainerModel>}
     * @memberof LokiBlobMetadataStore
     */
    private getContainerWithLeaseUpdated;
    /**
     * Get a container document from Loki collections.
     * Will throw ContainerNotFound error when container doesn't exist.
     *
     * @private
     * @param {string} account
     * @param {string} container
     * @param {Context} context
     * @param {true} [forceExist]
     * @returns {Promise<ContainerModel>}
     * @memberof LokiBlobMetadataStore
     */
    private getContainer;
    /**
     * Get a blob document model from Loki collection.
     * Will throw BlobNotFound storage error if blob doesn't exist.
     *
     * @private
     * @param {string} account
     * @param {string} container
     * @param {string} blob
     * @param {(string | undefined)} snapshot
     * @param {Context} context
     * @param {undefined} [forceExist]
     * @param {boolean} [forceCommitted] If true, will take uncommitted blob as a non-exist blob and throw exception.
     * @returns {Promise<BlobModel>}
     * @memberof LokiBlobMetadataStore
     */
    private getBlobWithLeaseUpdated;
    /**
     * Get the tier setting from request headers.
     *
     * @private
     * @param {string} tier
     * @returns {(Models.AccessTier | undefined)}
     * @memberof BlobHandler
     */
    private parseTier;
}
//# sourceMappingURL=LokiBlobMetadataStore.d.ts.map