import { Options as SequelizeOptions } from "sequelize";
import * as Models from "../generated/artifacts/models";
import Context from "../generated/Context";
import IBlobMetadataStore, { AcquireBlobLeaseResponse, AcquireContainerLeaseResponse, BlobId, BlobModel, BlockModel, BreakBlobLeaseResponse, BreakContainerLeaseResponse, ChangeBlobLeaseResponse, ChangeContainerLeaseResponse, ContainerModel, CreateSnapshotResponse, GetBlobPropertiesRes, GetContainerAccessPolicyResponse, GetContainerPropertiesResponse, GetPageRangeResponse, IContainerMetadata, IExtentChunk, ReleaseBlobLeaseResponse, RenewBlobLeaseResponse, RenewContainerLeaseResponse, ServicePropertiesModel, SetContainerAccessPolicyOptions } from "./IBlobMetadataStore";
/**
 * A SQL based Blob metadata storage implementation based on Sequelize.
 * Refer to CONTRIBUTION.md for how to setup SQL database environment.
 *
 * @export
 * @class SqlBlobMetadataStore
 * @implements {IBlobMetadataStore}
 */
export default class SqlBlobMetadataStore implements IBlobMetadataStore {
    private initialized;
    private closed;
    private readonly sequelize;
    /**
     * Creates an instance of SqlBlobMetadataStore.
     *
     * @param {string} connectionURI For example, "postgres://user:pass@example.com:5432/dbname"
     * @param {SequelizeOptions} [sequelizeOptions]
     * @memberof SqlBlobMetadataStore
     */
    constructor(connectionURI: string, sequelizeOptions?: SequelizeOptions);
    init(): Promise<void>;
    isInitialized(): boolean;
    close(): Promise<void>;
    isClosed(): boolean;
    clean(): Promise<void>;
    setServiceProperties(context: Context, serviceProperties: ServicePropertiesModel): Promise<ServicePropertiesModel>;
    getServiceProperties(context: Context, account: string): Promise<ServicePropertiesModel | undefined>;
    listContainers(context: Context, account: string, prefix: string | undefined, maxResults: number | undefined, marker: string): Promise<[ContainerModel[], string | undefined]>;
    createContainer(context: Context, container: ContainerModel): Promise<ContainerModel>;
    getContainerProperties(context: Context, account: string, container: string, leaseAccessConditions?: Models.LeaseAccessConditions): Promise<GetContainerPropertiesResponse>;
    deleteContainer(context: Context, account: string, container: string, options?: Models.ContainerDeleteMethodOptionalParams): Promise<void>;
    setContainerMetadata(context: Context, account: string, container: string, lastModified: Date, etag: string, metadata?: IContainerMetadata, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<void>;
    getContainerACL(context: Context, account: string, container: string, leaseAccessConditions?: Models.LeaseAccessConditions | undefined): Promise<GetContainerAccessPolicyResponse>;
    setContainerACL(context: Context, account: string, container: string, setAclModel: SetContainerAccessPolicyOptions): Promise<void>;
    acquireContainerLease(context: Context, account: string, container: string, options: Models.ContainerAcquireLeaseOptionalParams): Promise<AcquireContainerLeaseResponse>;
    releaseContainerLease(context: Context, account: string, container: string, leaseId: string, options?: Models.ContainerReleaseLeaseOptionalParams): Promise<Models.ContainerProperties>;
    renewContainerLease(context: Context, account: string, container: string, leaseId: string, options?: Models.ContainerRenewLeaseOptionalParams): Promise<RenewContainerLeaseResponse>;
    breakContainerLease(context: Context, account: string, container: string, breakPeriod: number | undefined, options?: Models.ContainerBreakLeaseOptionalParams): Promise<BreakContainerLeaseResponse>;
    changeContainerLease(context: Context, account: string, container: string, leaseId: string, proposedLeaseId: string, options?: Models.ContainerChangeLeaseOptionalParams): Promise<ChangeContainerLeaseResponse>;
    checkContainerExist(context: Context, account: string, container: string): Promise<void>;
    createBlob(context: Context, blob: BlobModel, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<void>;
    downloadBlob(context: Context, account: string, container: string, blob: string, snapshot?: string, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<BlobModel>;
    listBlobs(context: Context, account: string, container: string, blob?: string, prefix?: string, maxResults?: number, marker?: string, includeSnapshots?: boolean): Promise<[BlobModel[], any | undefined]>;
    listAllBlobs(maxResults?: number, marker?: string, includeSnapshots?: boolean): Promise<[BlobModel[], any | undefined]>;
    stageBlock(context: Context, block: BlockModel, leaseAccessConditions?: Models.LeaseAccessConditions): Promise<void>;
    getBlockList(context: Context, account: string, container: string, blob: string, isCommitted?: boolean, leaseAccessConditions?: Models.LeaseAccessConditions): Promise<any>;
    commitBlockList(context: Context, blob: BlobModel, blockList: {
        blockName: string;
        blockCommitType: string;
    }[], leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<void>;
    getBlobProperties(context: Context, account: string, container: string, blob: string, snapshot?: string, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<GetBlobPropertiesRes>;
    undeleteBlob(): Promise<void>;
    createSnapshot(context: Context, account: string, container: string, blob: string, leaseAccessConditions?: Models.LeaseAccessConditions, metadata?: Models.BlobMetadata, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<CreateSnapshotResponse>;
    deleteBlob(context: Context, account: string, container: string, blob: string, options: Models.BlobDeleteMethodOptionalParams): Promise<void>;
    setBlobHTTPHeaders(context: Context, account: string, container: string, blob: string, leaseAccessConditions: Models.LeaseAccessConditions | undefined, blobHTTPHeaders: Models.BlobHTTPHeaders | undefined, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<Models.BlobProperties>;
    setBlobMetadata(context: Context, account: string, container: string, blob: string, leaseAccessConditions: Models.LeaseAccessConditions | undefined, metadata: Models.BlobMetadata | undefined, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<Models.BlobProperties>;
    acquireBlobLease(context: Context, account: string, container: string, blob: string, duration: number, proposedLeaseId?: string, options?: Models.BlobAcquireLeaseOptionalParams): Promise<AcquireBlobLeaseResponse>;
    releaseBlobLease(context: Context, account: string, container: string, blob: string, leaseId: string, options?: Models.BlobReleaseLeaseOptionalParams): Promise<ReleaseBlobLeaseResponse>;
    renewBlobLease(context: Context, account: string, container: string, blob: string, leaseId: string, options?: Models.BlobRenewLeaseOptionalParams): Promise<RenewBlobLeaseResponse>;
    changeBlobLease(context: Context, account: string, container: string, blob: string, leaseId: string, proposedLeaseId: string, options?: Models.BlobChangeLeaseOptionalParams): Promise<ChangeBlobLeaseResponse>;
    breakBlobLease(context: Context, account: string, container: string, blob: string, breakPeriod: number | undefined, options?: Models.BlobBreakLeaseOptionalParams): Promise<BreakBlobLeaseResponse>;
    checkBlobExist(context: Context, account: string, container: string, blob: string, snapshot?: string | undefined): Promise<void>;
    getBlobType(account: string, container: string, blob: string, snapshot?: string | undefined): Promise<{
        blobType: Models.BlobType | undefined;
        isCommitted: boolean;
    } | undefined>;
    startCopyFromURL(context: Context, source: BlobId, destination: BlobId, copySource: string, metadata: Models.BlobMetadata | undefined, tier: Models.AccessTier | undefined): Promise<Models.BlobProperties>;
    setTier(context: Context, account: string, container: string, blob: string, tier: Models.AccessTier, leaseAccessConditions?: Models.LeaseAccessConditions): Promise<200 | 202>;
    uploadPages(context: Context, blob: BlobModel, start: number, end: number, persistency: IExtentChunk, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<Models.BlobProperties>;
    clearRange(context: Context, blob: BlobModel, start: number, end: number, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<Models.BlobProperties>;
    getPageRanges(context: Context, account: string, container: string, blob: string, snapshot?: string | undefined, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<GetPageRangeResponse>;
    resizePageBlob(context: Context, account: string, container: string, blob: string, blobContentLength: number, leaseAccessConditions?: Models.LeaseAccessConditions, modifiedAccessConditions?: Models.ModifiedAccessConditions): Promise<Models.BlobProperties>;
    updateSequenceNumber(context: Context, account: string, container: string, blob: string, sequenceNumberAction: Models.SequenceNumberActionType, blobSequenceNumber: number | undefined): Promise<Models.BlobProperties>;
    appendBlock(context: Context, block: BlockModel, leaseAccessConditions?: Models.LeaseAccessConditions | undefined, modifiedAccessConditions?: Models.ModifiedAccessConditions | undefined, appendPositionAccessConditions?: Models.AppendPositionAccessConditions | undefined): Promise<Models.BlobProperties>;
    listUncommittedBlockPersistencyChunks(marker?: string, maxResults?: number): Promise<[IExtentChunk[], string | undefined]>;
    iteratorExtents(): AsyncIterator<string[]>;
    private getModelValue;
    private deserializeModelValue;
    private serializeModelValue;
    /**
     * This method will restore object to Uint8Array.
     *
     * @private
     * @param {*} obj
     * @returns {(Uint8Array | undefined)}
     * @memberof LokiBlobMetadataStore
     */
    private restoreUint8Array;
    private convertDbModelToContainerModel;
    private convertContainerModelToDbModel;
    private convertDbModelToBlobModel;
    private convertBlobModelToDbModel;
    private convertDbModelToBlobContentProperties;
    private convertBlobContentPropertiesToDbModel;
    private convertDbModelToLease;
    private convertLeaseToDbModel;
}
//# sourceMappingURL=SqlBlobMetadataStore.d.ts.map