/// <reference types="node" />
import { SASProtocol } from "./IAccountSASSignatureValues";
import { IIPRange } from "./IIPRange";
/**
 * ONLY AVAILABLE IN NODE.JS RUNTIME.
 *
 * IQueueSASSignatureValues is used to help generating Queue service SAS tokens for queues.
 *
 * @export
 * @class IQueueSASSignatureValues
 */
export interface IQueueSASSignatureValues {
    /**
     * The version of the service this SAS will target. If not specified, it will default to the version targeted by the
     * library.
     *
     * @type {string}
     * @memberof IQueueSASSignatureValues
     */
    version?: string;
    /**
     * Optional. SAS protocols, HTTPS only or HTTPSandHTTP
     *
     * @type {SASProtocol | string}
     * @memberof IQueueSASSignatureValues
     */
    protocol?: SASProtocol | string;
    /**
     * Optional. When the SAS will take effect.
     *
     * @type {Date | string}
     * @memberof IQueueSASSignatureValues
     */
    startTime?: Date | string;
    /**
     * Optional only when identifier is provided. The time after which the SAS will no longer work.
     *
     * @type {Date | string}
     * @memberof IQueueSASSignatureValues
     */
    expiryTime?: Date | string;
    /**
     * Optional only when identifier is provided.
     * Please refer to {@link QueueSASPermissions}
     * being accessed for help constructing the permissions string.
     *
     * @type {string}
     * @memberof IQueueSASSignatureValues
     */
    permissions?: string;
    /**
     * Optional. IP ranges allowed in this SAS.
     *
     * @type {IIPRange | string}
     * @memberof IQueueSASSignatureValues
     */
    ipRange?: IIPRange | string;
    /**
     * The name of the queue the SAS user may access.
     *
     * @type {string}
     * @memberof IQueueSASSignatureValues
     */
    queueName: string;
    /**
     * Optional. The name of the access policy on the queue this SAS references if any.
     *
     * @see https://docs.microsoft.com/en-us/rest/api/storageservices/establishing-a-stored-access-policy
     *
     * @type {string}
     * @memberof IQueueSASSignatureValues
     */
    identifier?: string;
}
/**
 * Creates an instance of SASQueryParameters.
 *
 * Only accepts required settings needed to create a SAS. For optional settings please
 * set corresponding properties directly, such as permissions, startTime and identifier.
 *
 * WARNING: When identifier is not provided, permissions and expiryTime are required.
 * You MUST assign value to identifier or expiryTime & permissions manually if you initial with
 * this constructor.
 *
 * @export
 * @param {IQueueSASSignatureValues} queueSASSignatureValues
 * @param {SharedKeyCredential} sharedKeyCredential
 * @returns {SASQueryParameters}
 */
export declare function generateQueueSASSignature(queueSASSignatureValues: IQueueSASSignatureValues, accountName: string, sharedKey: Buffer): [string, string];
//# sourceMappingURL=IQueueSASSignatureValues.d.ts.map