import Context from "../generated/Context";
import { QUEUE_STATUSCODE } from "../utils/constants";
import { IQueueMetadata, IQueueMetadataStore, MessageModel, MessageUpdateProperties, QueueACL, QueueModel, ServicePropertiesModel } from "./IQueueMetadataStore";
/**
 * This is a metadata source implementation for queue based on loki DB.
 *
 * Notice that, following design is for emulator purpose only, and doesn't design for best performance.
 * We may want to optimize the persistency layer performance in the future. Such as by distributing metadata
 * into different collections, or make binary payload write as an append-only pattern.
 *
 * Loki DB includes following collections and documents:
 *
 * -- SERVICE_PROPERTIES_COLLECTION // Collection contains service properties
 *                                  // Default collection name is $SERVICES_COLLECTION$
 *                                  // Each document maps to 1 account queue service
 *                                  // Unique document properties: accountName
 * -- QUEUES_COLLECTION  // Collection contains all queues
 *                           // Default collection name is $QUEUES_COLLECTION$
 *                           // Each document maps to 1 queue
 *                           // Unique document properties: accountName, (queue)name
 * -- MESSAGES_COLLECTION    // Collection contains all messages
 *                           // Default collection name is $MESSAGES_COLLECTION$
 *                           // Each document maps to a message
 *                           // Unique document properties: accountName, queueName, messageId
 *
 * @export
 * @class LokiQueueMetadataStore
 */
export default class LokiQueueMetadataStore implements IQueueMetadataStore {
    readonly lokiDBPath: string;
    private readonly db;
    private initialized;
    private closed;
    private readonly SERVICES_COLLECTION;
    private readonly QUEUES_COLLECTION;
    private readonly MESSAGES_COLLECTION;
    constructor(lokiDBPath: string);
    isInitialized(): boolean;
    isClosed(): boolean;
    init(): Promise<void>;
    /**
     * Close loki DB.
     *
     * @returns {Promise<void>}
     * @memberof LokiQueueDataStore
     */
    close(): Promise<void>;
    /**
     * Update queue service properties. Create service properties document if not exists in persistency layer.
     * Assume service properties collection has been created during start method.
     *
     * @param {ServicePropertiesModel} updateProperties
     * @returns {Promise<void>}
     * @memberof LokiQueueMetadataStore
     */
    updateServiceProperties(updateProperties: ServicePropertiesModel): Promise<void>;
    /**
     * Get service properties for specific storage account.
     *
     * @template T
     * @param {string} account
     * @returns {(Promise<ServicePropertiesModel | undefined>)}
     * @memberof LokiQueueMetadataStore
     */
    getServiceProperties(account: string): Promise<ServicePropertiesModel | undefined>;
    /**
     * List queues with query conditions specified.
     *
     * @template T
     * @param {string} account
     * @param {string} [prefix=""]
     * @param {number} [maxResults=5000]
     * @param {number} [marker=0]
     * @returns {(Promise<[QueueModel[], number | undefined]>)} A tuple including queues and next marker
     * @memberof LokiQueueMetadataStore
     */
    listQueues(account: string, prefix?: string, maxResults?: number, marker?: number): Promise<[QueueModel[], number | undefined]>;
    /**
     * Get a queue item from persistency layer by account and queue name.
     *
     * @param {string} account
     * @param {string} queue
     * @param {Context} [context]
     * @returns {Promise<QueueModel>}
     * @memberof LokiQueueMetadataStore
     */
    getQueue(account: string, queue: string, context?: Context): Promise<QueueModel>;
    /**
     * Create a queue in persistency layer.
     * Return 201 if create a new one, 204 if a same one exist, 409 error if a conflit one exist.
     * @see https://docs.microsoft.com/en-us/rest/api/storageservices/create-queue4
     *
     * @param {QueueModel} queue
     * @param {Context} [context]
     * @returns {Promise<QUEUE_STATUSCODE>}
     * @memberof LokiQueueMetadataStore
     */
    createQueue(queue: QueueModel, context?: Context): Promise<QUEUE_STATUSCODE>;
    /**
     * Delete a queue and its all messages.
     *
     * @param {string} account
     * @param {string} queue
     * @param {Context} [context]
     * @returns {Promise<void>}
     * @memberof LokiQueueMetadataStore
     */
    deleteQueue(account: string, queue: string, context?: Context): Promise<void>;
    /**
     * Update the ACL of an exist queue item in persistency layer.
     *
     * @param {string} account
     * @param {string} queue
     * @param {QueueACL} [queueACL]
     * @param {Context} [context]
     * @returns {Promise<void>}
     * @memberof LokiQueueMetadataStore
     */
    setQueueACL(account: string, queue: string, queueACL?: QueueACL, context?: Context): Promise<void>;
    /**
     * Update the metadata of an exist queue item in persistency layer.
     *
     * @param {string} account
     * @param {string} queue
     * @param {QueueMetadata} [metadata]
     * @param {string} [requestId]
     * @returns {Promise<void>}
     * @memberof LokiQueueMetadataStore
     */
    setQueueMetadata(account: string, queue: string, metadata?: IQueueMetadata, context?: Context): Promise<void>;
    /**
     * Get the number of messages of a queue from persistency layer.
     *
     * @param {string} account
     * @param {string} queue
     * @param {string} [requestId]
     * @returns {number}
     * @memberof LokiQueueMetadataStore
     */
    getMessagesCount(account: string, queue: string, context?: Context): Promise<number>;
    /**
     * Insert a message in metadata.
     * The existence of the certain queue should be validated before.
     *
     * @param {MessageModel} message
     * @param {Context} [context]
     * @returns {Promise<void>}
     * @memberof LokiQueueMetadataStore
     */
    insertMessage(message: MessageModel, context?: Context): Promise<void>;
    /**
     * peek messages of a given number in persistency layer.
     *
     * @param {string} account
     * @param {string} queue
     * @param {number} [numOfMessages]
     * @param {Date} [queryDate]
     * @param {Context} [context]
     * @returns {Promise<MessageModel[]>}
     * @memberof LokiQueueMetadataStore
     */
    peekMessages(account: string, queue: string, numOfMessages?: number, queryDate?: Date, context?: Context): Promise<MessageModel[]>;
    /**
     * Dequeue messages from a queue in persistency layer.
     *
     * @param {string} account
     * @param {string} queue
     * @param {Date} timeNextVisible
     * @param {string} popReceipt
     * @param {number} [numOfMessages]
     * @param {Date} [queryDate]
     * @param {Context} [context]
     * @returns {Promise<MessageModel[]>}
     * @memberof LokiQueueMetadataStore
     */
    getMessages(account: string, queue: string, timeNextVisible: Date, popReceipt: string, numOfMessages?: number, queryDate?: Date, context?: Context): Promise<MessageModel[]>;
    /**
     * Delete the metadata of a message from persistency layer.
     *
     * @param {string} account
     * @param {string} queue
     * @param {string} messageId
     * @param {string} validatingPopReceipt
     * @param {Context} [context]
     * @returns {Promise<void>}
     * @memberof LokiQueueMetadataStore
     */
    deleteMessage(account: string, queue: string, messageId: string, validatingPopReceipt: string, context?: Context): Promise<void>;
    /**
     * Update the metadata of an exist message in persistency layer.
     *
     * @param {MessageUpdateProperties} message
     * @param {string} validatingPopReceipt
     * @param {Context} [context]
     * @returns {Promise<void>}
     * @memberof LokiQueueMetadataStore
     */
    updateMessage(message: MessageUpdateProperties, validatingPopReceipt: string, context?: Context): Promise<void>;
    /**
     * Clear the metadata of all messages in a given queue from persistency layer.
     *
     * @param {string} account
     * @param {string} queue
     * @param {Context} [context]
     * @returns {Promise<void>}
     * @memberof LokiQueueMetadataStore
     */
    clearMessages(account: string, queue: string, context?: Context): Promise<void>;
    /**
     * List messages.
     *
     * @param {number} [maxResults]
     * @param {(number | undefined)} [marker]
     * @returns {(Promise<[MessageModel[], number | undefined]>)}
     * @memberof IQueueMetadataStore
     */
    listMessages(maxResults?: number, marker?: number | undefined): Promise<[MessageModel[], number | undefined]>;
    iteratorExtents(): AsyncIterator<string[]>;
    /**
     * Check the existence of a queue.
     *
     * @private
     * @param {string} account
     * @param {string} queue
     * @param {Context} [context]
     * @memberof LokiQueueMetadataStore
     */
    private checkQueueExist;
    /**
     * Deep copy a message, return the new object.
     *
     * @private
     * @param {MessageModel} message
     * @returns {MessageModel}
     * @memberof LokiQueueMetadataStore
     */
    private messageCopy;
    /**
     *  Deep copy a queue, return the new object.
     *
     * @private
     * @param {QueueModel} queue
     * @returns {QueueModel}
     * @memberof LokiQueueMetadataStore
     */
    private queueCopy;
    /**
     * Escape a string to be used as a regex.
     *
     * @private
     * @param {string} regex
     * @returns {string}
     * @memberof LokiQueueMetadataStore
     */
    private escapeRegex;
}
//# sourceMappingURL=LokiQueueMetadataStore.d.ts.map